#!python

#Joe Agajanian and Taylor Brent

# Magic The Gathering AI
# This is an application of Monte Carlo
# Tree search to the game of Magic



import copy
import random
from heapq import *

# these are the global variables
gameOver = False
mainPlayList = []
attackList = []
blockList = []

primaryHeap = []
opponentHeap = []
opponentDictionary = {}

gameOverMessage = ""
turn = 0;
dictOfCards = {}
decks = []
# these are the keywords that creatures can have 
HASTE = "haste"
TRAMPLE = "trample"
FLYING = "flying"
VIGILANCE = "vigilance"
INDESTRUCTIBLE = "indestructible"
LIFELINK = "lifelink"
FIRSTSTRIKE="firststrike"
DOUBLESTRIKE="doublestrike"
# these are the 5 different colors a creature can have
WHITE = "W"
RED = "R"
BLACK = "B"
BLUE = "U"
GREEN = "G"
COLORLESS = "*"
COLORS = [WHITE, RED, BLACK, BLUE, GREEN]
# these are the different types of players (random and monte carlo tree searchers)
TYPERANDOM = "TYPERANDOM"
TYPEMCTS = "TYPEMCTS"


class Card:
# Parent class for all card types
    def __init__(self):
        print "Why are you creating a card you should not be creating a card"
        #This should never happen

class Creature(Card):
    def __init__(self, name, cost, keywords, colors, power, toughness):
        self.name = name
        self.cost = cost
        self.keywords = keywords
        self.power = power
        self.toughness = toughness
        self.colors = colors
        self.tapped = False
        self.damage = 0
        self.powerBuff = 0
        self.toughnessBuff = 0
        self.sick = HASTE not in keywords
        self.indestructible = INDESTRUCTIBLE in keywords
        self.isPermanent = True
        self.isBlocking = False

    def copy(self):
        newCreature = Creature(self.name, self.cost, self.keywords, self.colors, self.power,self.toughness)
        return newCreature

    def getName(self):
            return self.name
    def getKeywords(self):
            return self.keywords
    def getPower(self):
            return self.power + self.powerBuff
    def getOriginalPower(self):
            return self.Power
    def getToughness(self):
            return self.toughness + self.toughnessBuff
    def getOriginalToughness(self):
            return self.toughness
    def getBlocking(self):
            return self.isBlocking
    def getCost(self):
            return len(self.cost)
    def isTapped(self):
            return self.tapped
    def isDead(self):
            return self.getToughness() <= 0
    def isPermanent(self):
            return self.isPermanent
    def isLand(self):
            return False
    def tap(self):
        self.tapped = True
    def untap(self):
        self.tapped = False
    def doDamage(self, damage):
        self.damage += damage
    def buffPower(self, buff):
        self.powerBuff = buff
    def buffToughness(self, buff):
        self.toughnessBuff = buff
    def buffCreature(self, powerBuff, toughnessBuff):
        self.powerBuff += powerBuff
        self.toughnessBuff += toughnessBuff
    def blocking(self):
        self.isBlocking = True
    def clear(self):
        self.sick = False
        self.tapped = False
        self.damage = 0
        self.powerBuff = 0
        self.toughnessBuff = 0
        self.isBlocking = False

    def endTurn(self):
        self.powerBuff = 0
        self.toughnessBuff = 0
        self.damage = 0
        self.isBlocking = False
    

    def __str__(self):
        costRep = ""
        for elem in self.cost:
            costRep += elem
        string = "Name: " + self.name + " - " + costRep + "\n"
        string += "Cost: " + str(self.getCost()) + "\n"
        string += "Keywords: "
        for keyword in self.keywords:
            string += keyword + " "
        string += "\nAttributes: "
        string += str(self.power) + "/" + str(self.toughness) + "\nStatus: "
        if self.tapped:
            string += "Tapped \n"
        else:
            string += "Untapped \n"
        return string
        
    

class Land(Card):
    def __init__(self, name, colors):
        self.name = name
        self.colors = colors
        self.tapped = False
        self.isPermanent = True

    def copy(self):
        newLand = Land(self.name, self.colors)
        return newLand

    def getName(self):
        return self.name

    def tap(self):
        if self.tapped == False:
            self.tapped = True
            return True
        else:
            return False

    def getCost(self):
        return 0

    def untap(self):
        self.tapped=False

    def isTapped(self):
        return self.tapped

    def clear(self):
        self.tapped = False

    def isPermanent(self):
        return self.isPermanent

    def isLand(self):
        return True

    def __str__(self):
        string = "Name: " + self.name + "\nColors: "
        for color in self.colors:
            string += color
        string += "\nStatus: "
        if self.tapped:
            string += "Tapped \n"
        else:
            string += "Untapped \n"
        return string

class Deck:
    def __init__(self, name):
        # These are the Decks
        self.name = name
        self.cardList = []

    def takeOffTop(self):
        return self.pullCard(0)

    def pullCard(self, index):
        card = self.cardList[index]
        self.cardList = self.cardList[:index] + self.cardList[(index+1):]
        return card

    def size(self):
        return len(self.cardList)

    # used to make the deck only
    def addCard(self, card, number):
        for i in range(0,number):
            tempCard = card.copy()
            self.cardList.append(tempCard)

    def getName(self):
        return self.name

    def shuffle(self):
        random.shuffle(self.cardList)


class Player:
    global gameOver
    def __init__(self, deck, playerType):
        self.lands = []
        self.creatures = []
        self.hand = []
        self.deck = deck
        self.life = 20
        self.type = playerType
        self.winCount = 0

    def copy(self):
        newPlayer = Player(self.deck)
        newPlayer.lands = self.lands
        newPlayer.creatures = self.creatures
        newPlayer.hand = self.hand
        newPlayer.deck = self.deck
        newPlayer.life = self.life
        newPlayer.type = self.type
        newPlayer.winCount = self.winCount
        return newPlayer


    def clearWinCount(self):
        self.winCount = 0
    def getWinCount(self):
        return self.winCount
    def incrementWinCount(self):
        self.winCount += 1
    def getType(self):
        return self.type
    def drawHand(self):
        global gameOver
        for i in range(0, 7):
            self.draw()
    def getDeck(self):
        return self.deck
    def getName(self):
        return self.deck.getName()
    def getLife(self):
        return self.life
    def getHand(self):
        return self.hand
    def getCreatures(self):
        return self.creatures
    def getLands(self):
        return self.lands
    def getBlockList(self, creature):
        return self.blockersInPlay(creature)
    def getMonteBlockList(self, creature):
        cList = self.blockersInPlay(creature)
        cList.append(0)
        return cList
    def findCreature(self, creature):
        name = creature.getName()
        for card in self.creatures:
            if card.getName() == name:
                return card
        print "FATAL ERROR"
    
    # gets a semi-random blocker for a creature
    def quasiBlockersInPlay(self, creature):
        creatures = []
        for card in self.creatures:
            if (not card.isTapped()):
                if (not card.getBlocking()):
                    if (FLYING in creature.getKeywords()):
                        if (FLYING in card.getKeywords()):
                            creatures.append(card)
                    else:
                        creatures.append(card)
        if creatures != []:
            return random.choice(creatures)
        else:
            return 0
    # gets a semi-random list of attackers
    def quasiAttackList(self):
        finalList = []
        aList = self.possibleAttackersInPlay()
        for pAttcker in aList:
            if random.random() < .5:
                finalList.append(pAttcker)
        return finalList

    # gets a semi-random list of creatures to play
    def quasiPlayList(self):
        finalList = []
        mana = self.getNumUntappedLands()
        aList = self.CreaturesInHand()
        for pAttcker in aList:
            if pAttcker.getCost() <= mana:
                if random.random() < .5:
                    mana -= pAttcker.getCost()
                    finalList.append(pAttcker)
        return finalList

    """
    this function was taken from 
    http://www.technomancy.org/python/powerset-generator-python/
    to improve the runtime and memory of the powerset function 
    that we wrote.
    While it is not used, it inspired a better solution
    """
    def powerSetAttackersGenerator(self, attackList):
        if len(attackList) < 1:
            yield attackList
            # yield []
        else:
            for attacker in self.powerSetAttackersGenerator(attackList[1:]):
                yield [attackList[0]] + attacker
                yield attacker
    """
    this function was taken from 
    http://www.technomancy.org/python/powerset-generator-python/
    to improve the runtime and memory of the powerset function 
    that we wrote.
    While it is not used, it inspired a better solution
    """        
    def powerSetCreaturesGenerator(self, cList, mana):
        sum1 = 0
        sum2 = 0
        if len(cList) < 1:
            yield cList
            # yield []
        else:
            for creatureList in self.powerSetCreaturesGenerator(cList[1:], mana):
                sum1 = 0
                sum2 = 0
                for creature in cList:
                    sum1 += creature.getCost()
                for creature in creatureList:
                    sum2 += creature.getCost()
                if (sum1 + sum2) <= mana:
                    yield [cList[0]] + creatureList
                if sum1 <= mana:
                    yield cList



    # gets all possible combinations of attackers (exponential runtime)
    def getAttackList(self):
        global attackList
        possibleAttackers = self.attackersInPlay()
        attackList = []
        self.getAttackHelper(possibleAttackers, [])
        return attackList

    # gets all creatures able to attack
    def possibleAttackersInPlay(self):
        creatures = []
        for card in self.creatures:
            if (not card.sick):
                if (not card.isTapped()):
                    creatures.append(card)
        return creatures

    # gets all creatures that can be played (only a subset of those returned can be played though)
    def getPlayableCreatures(self):
        global mainPlayList 
        # main = []
        mana = copy.deepcopy(self.getNumUntappedLands())
        creatureList = copy.deepcopy(self.CreaturesInHand())
        mainPlayList = []
        self.getPlayableCreaturesHelper(mana, creatureList, [])
        return mainPlayList

    # gets a list of creatures that can attack
    def attackersInPlay(self):
        creatures = []
        for card in self.creatures:
            if (not card.sick):
                if (not card.isTapped()):
                    creatures.append(card)
        return creatures

    # used to pay to play creatures in hand
    def tapLand(self):
        for land in self.lands:
            result = land.tap()
            if result:
                return True
        return False

    # as a result of battle, players lost life by taking damage
    def takeDamage(self, damage):
        self.life -= damage

    def draw(self):
        global gameOver
        if self.deck.size() >=1:
            self.hand.append(self.deck.takeOffTop())
        else:
            gameOver = True
            gameOverMessage = "Player " + self.deck.getName() + " lost"

    def playLand(self, land):
        self.lands.append(land)
        self.hand.remove(land)

    def playCreatures(self, creatures):
        for creature in creatures:
            self.playCreature(creature)

    # handles paying for the creatures that are being played
    # returns whether or not the playing was successful
    def playCreature(self, creature):
        success = False
        cost = creature.getCost()
        untappedLands = self.getNumUntappedLands()
        if cost <= untappedLands:
            success = True
            for i in range(0, cost):
                self.tapLand()
            self.creatures.append(creature)
            for card in self.hand:
                if card.getName() == creature.getName():
                    self.hand.remove(card)
                    return success

                
        return success

    # creatures die as a result of battle and are removed in this function
    def kill(self, creature):
        for card in self.creatures:
            if card.getName() == creature.getName():
                self.creatures.remove(card)
                return
        print "FATAL ERROR"

    # determines whether or not the player has lost
    def isDead(self):
        if self.life <= 0 or self.deck.size() == 0:
            return True
        else:
            return False

    # at the beginning of each turn, we clear 'summoning sickness*' 
    # and damage done during previous battles along with untapping creatures
    # *see MTG for rules and definition of summoning sickness
    def clearAll(self):
        # Warning, clears summoning sickness
        for card in self.creatures:
            card.clear()
        for card in self.lands:
            card.clear()

    # at the end of each turn, the player attacking switches
    def endTurn(self):
        for card in self.creatures:
            card.endTurn()

    def shuffleDeck(self):
        self.deck.shuffle()

    def getNumUntappedLands(self):
        i = 0
        for land in self.lands:
            if not land.isTapped():
                i+=1
        return i

    def getNumTappedLands(self):
        i = 0
        for land in self.lands:
            if land.isTapped():
                i+=1
        return i
    # return the set of all creatures that have the ability to block a creature
    # only one (or none) will actually do the blocking
    def blockersInPlay(self, creature):
        creatures = []
        for card in self.creatures:
            if (not card.isTapped()):
                if (not card.getBlocking()):
                    if (FLYING in creature.getKeywords()):
                        if (FLYING in card.getKeywords()):
                            creatures.append(card)
                    else:
                        creatures.append(card)
        return creatures

    # gets a list of creatures in hand
    def CreaturesInHand(self):
        creatures = []
        for card in self.hand:
            if (not card.isLand()):
                creatures.append(card)
        return creatures

    # helper to get all possible attackers
    def getAttackHelper(self, creatureList, possibleAttackers):
        global attackList
        # if len(creatureList)
        if (creatureList == []):
            if (possibleAttackers != []):
                if (possibleAttackers not in attackList):
                    attackList.append(possibleAttackers)
            return
        for card in creatureList:
            loseIt = self.getAttackHelper(creatureList[1:], possibleAttackers)
            cList = copy.deepcopy(possibleAttackers)
            cList.append(card)
            creatureList.remove(card)
            useit = self.getAttackHelper(creatureList, cList)
        return

    # helper function to get all possible combinations of
    # creatures that can be play (exponential runtime)
    def getPlayableCreaturesHelper(self, mana, creatureList, playableList):
        global mainPlayList
        # if len(creatureList)
        if (creatureList == []):
            if (playableList != []):
                if (playableList not in mainPlayList):
                    mainPlayList.append(playableList)
            return
        if (mana == 0):
            if (playableList != []):
                if (playableList not in mainPlayList):
                    mainPlayList.append(playableList)
            return

        for card in creatureList:
            loseIt = self.getPlayableCreaturesHelper(mana, creatureList[1:], playableList)
            if (card.getCost() <= mana):
                cList = copy.deepcopy(playableList)
                cList.append(card)
                creatureList.remove(card)
                useit = self.getPlayableCreaturesHelper(mana - card.getCost(), creatureList, cList)
        return

    # helper function to get all possible combinations of
    # creatures that can be play with a better runtime that the function above, but with less 
    # randomness to save memory and increase runtime
    def powerPlayableList(self, mana, creatureList, playableList):
        if ((creatureList == []) or (mana == 0)):
            creatureList = []
            yield playableList

        for card in creatureList:
            yield self.powerPlayableList(mana, creatureList[1:], playableList)
            if (card.getCost() <= mana):
                cList = copy.deepcopy(playableList)
                cList.append(card)
                creatureList.remove(card)
                yield self.powerPlayableList(mana - card.getCost(), creatureList, cList)


    def __str__(self):
        result = ""
        result += "Player: " + self.deck.getName() + "\n"
        result += "Decksize: " + str(self.deck.size()) + "\n"
        result += "Life: " + str(self.getLife()) + "\n"
        result += "Hand: ["
        for card in self.hand:
            result += str(card.getName()) + " - "
        result += "]\n"
        result += "Untapped lands: " + str(self.getNumUntappedLands()) + "\n"
        result +="Tapped lands: " + str(self.getNumTappedLands()) + "\n"
        result += "Creatures: " + "\n"
        for creature in self.creatures:
            result += str(creature) + "\n\n"
        result += ""
        return result


class Board:
    global gameOver
    def __init__(self, player1, player2):
        self.players = [player1, player2]

    def copy(self):
        newBoard = Board(self.players[0], self.players[1])
        return newBoard

    def getPlayer(self, index):
        return self.players[index]

    def shufflePlayerDeck(self, index):
        self.players[index].shuffleDeck()

    def __str__(self):
        return str(self.players[0]) + str(self.players[1])


class Game:
    global gameOver
    def __init__(self, player1, player2):
        self.board = Board(player1, player2)
        self.activePlayer = 0
        self.turn = 0
        self.board.shufflePlayerDeck(0)
        self.board.shufflePlayerDeck(1)
        self.board.getPlayer(0).drawHand()
        self.board.getPlayer(1).drawHand()
    def doTurn(self):

        current = self.board.getPlayer(self.activePlayer)
        # "Untap and Draw" phase
        current.clearAll()
        if self.turn != 0:
            current.draw()
        for card in current.getHand():
            if card.isLand():
                current.playLand(card)
                break
        # Main, Attack, and Block phase
        self.doAI(self.activePlayer, (self.activePlayer + 1) % 2)
        print self.board

    # checks to see if a player has lost the game 
    def checkOver(self):
        global gameOver
        global player1WinCount
        global player2WinCount        
        if self.board.getPlayer(0).isDead():
            gameOverMessage = "Player " + self.board.getPlayer(0).getName() + " lost"
            gameOver = True
            player2WinCount += 1
        elif self.board.getPlayer(1).isDead():
            gameOverMessage = "Player " + self.board.getPlayer(1).getName() + " lost"
            gameOver = True
            player1WinCount += 1

    # handles moving to the next turn and takes care of all 'cleaning up' of the board 
    def advanceTurn(self):
        current = self.board.getPlayer(self.activePlayer)
        current.endTurn()
        self.activePlayer = (self.activePlayer + 1) % 2
        self.board.getPlayer(self.activePlayer).endTurn()
        self.turn += 1
    def __str__(self):
        return str(self.board)

    # boardState heuristic:
    # the heuristic favors a large ratio of the players life against the opponents
    # along with a large power and toughness sum of all creatures compared to opponents
    # creatures.
    # this function can be used to cut off MCTS due to exponential runtime of MCTS
    def boardHeuristic(self, primary, opponent):
        lifeValue = 0
        if (opponent.getLife() != 0):
            lifeValue += 1.0*primary.getLife()/(opponent.getLife())
        else:
            lifeValue = 99999999
         
        creaturePoints = 0
        for card in primary.getCreatures():
            creaturePoints += (card.getOriginalPower() + card.getOriginalToughness())/2.0 + len(card.getKeywords())
        for opponentCard in opponent.getCreatures():
            creaturePoints -= (opponentCard.getOriginalPower() + opponentCard.getOriginalToughness())/2.0 + len(opponentCard.getKeywords())
        return creaturePoints + lifeValue

    # given a board, both players, an attacker and a blocker, damage is dealt according
    # to the rules of MTG* to both the creatures and potentially the players
    # resulting in possible loss of creatures and (possible loss or gain) life depending on the 
    # abilities of the pair of creatues. 
    # see MTG rules of battle for more info.

    def doBattle(self, board, primaryPlayer, attacker, opponent, blocker, simulation):
        attacker = board.getPlayer(primaryPlayer).findCreature(attacker)
        attacker.tap()
        if (blocker == 0):
            board.getPlayer(opponent).takeDamage(attacker.getPower())
        else:
            blocker = board.getPlayer(opponent).findCreature(blocker)
            blocker.blocking()
            blocker.buffCreature(0, -1 * attacker.getPower())
            attacker.buffCreature(0, -1 * blocker.getPower())
            if blocker.isDead():
                board.getPlayer(opponent).kill(blocker)
            if attacker.isDead():
                board.getPlayer(primaryPlayer).kill(attacker)
        return board
        
    # this function handles the random player's playing creature phase, and attempts
    # to reduce the runtime due to exponential branching along with
    # possible depth of 60    
    def randomPlayCreatures(self, playableList):
        if len(playableList) > 0:
            return random.choice(playableList)
        else:
            return []

    # this function handles the random player's attack phase, and attempts
    # to reduce the runtime due to exponential branching along with
    # possible depth of 60        
    def randomAttackPhase(self, attackingList):
        if len(attackingList) > 0:
            attacking = random.choice(attackingList)
            return attacking
        else:
            return []

    # this function handles the random player's block phase, and attempts
    # to reduce the runtime due to exponential branching along with
    # possible depth of 60        
    def randomBlockPhase(self, blockList):
        if len(blockList) > 0:
            blockList.append(0)
            blocking = random.choice(blockList)

            return blocking
        else:
            return []

    # this function handles the random player's Main phase, and attempts
    # to reduce the runtime due to exponential branching along with
    # possible depth of 60        
    def randomMainPhase(self, player):
        possiblePlayerList = player.getPlayableCreatures()
        playList = self.randomPlayCreatures(possiblePlayerList)
        # to speed up runtime, a substitute for this function is used.
        # while the speed is now increased, the randomness is only semi-random
        # playList = self.randomPowerCreature(player, player.CreaturesInHand())
        # cont = raw_input("see cards being played")
        for index in range(len(playList)):
            # print str(card)
            player.playCreature(playList[index])



    # this random attacker chooser helps to choose quickly and semi-randomly
    # a list of atters.
    # this is very memory intensive
    def randomAttacker(self, player, attackList):
        aList = list(player.powerSetAttackersGenerator(attackList))
        if aList == []:
            return []
        else:    
            return random.choice(aList)


    """
    the try except statement in this function was taken from:
    http://www.ibm.com/developerworks/library/l-pycon/index.html 
    to speed up runtime, a substitute for this function is no longer used.
    while the speed is now increased, the randomness is only semi-random
    so it is left here for potential paralellising
    """
    def randomPowerCreature(self, player, creaturesInHandList):
        cList = list(player.powerPlayableList(player.getNumUntappedLands(), creaturesInHandList, []))
        print "clist:             "
        finalList = []
        choice = random.choice(cList)
        try:
            while 1: 
                x = choice.next()
                finalList.append(x)
        except StopIteration:
            return finalList
      
    # this function generates all possible outcomes of a boardstate fiven a list of attackers 
    def getPossibleBoardsForAttackers(self, primaryPlayer, opponentPlayer, copyAttackers):
        curBoards = [self.board]
        attackers = copy.deepcopy(copyAttackers)
        while len(attackers) > 0:
            attacker = attackers.pop()
            nextBoards = []
            for possibleBoard in curBoards:
                blockList = possibleBoard.getPlayer(opponentPlayer).getMonteBlockList(attacker)
                for blocker in blockList:
                    boardCopy = copy.deepcopy(possibleBoard)
                    boardCopy = self.doBattle(boardCopy, primaryPlayer,
                                                         attacker, opponentPlayer, blocker, False)
                    nextBoards.append(boardCopy)
            curBoards = nextBoards
        return curBoards

    # if the player is random, this fucntion calls the appropriate helperfunctions for the random player
    def RandomPlayAndAttack(self, primary, opponent):
        self.board = self.RandomPlay(primary, opponent)
        attackers = self.RandomAttack(primary, opponent)
        return attackers

    # if the player is using MCTS, we call this function to do all playing and blocking 
    # if it is the appropriate players turn
    def MCTSPlayAndAttack(self, primary, opponent):
        # due to the exponential runtime, the runtime was reduced by making the first part of move 
        # the first move randomly, but play quality is only slightly reduced
        self.board = self.RandomPlay(primary, opponent) # CHANGE TO MCTSPLAY EVENTUALLY
        attackers = self.MCTSAttack(primary, opponent)
        return attackers

    # this function prints the results before combat (think of it as a battle)
    # and alse the result of the battle 
    def MCTSAttack(self, primary, opponent):
        attackList = self.board.getPlayer(primary).getAttackList()
        attackList.append([])
        print "Before combat:"
        print self.board
        heuristicsAndAttackers = []
        print "Combat just happened. Here's the result:\n"
        if len(attackList) > 0:
            for attackers in attackList:
                possibleBoards = self.getPossibleBoardsForAttackers(primary, opponent, attackers)
                # Run random MCTS simulation to get heuristic
                [board, heuristic] = self.getHeuristicMCTS(primary, opponent, possibleBoards)
                heappush(heuristicsAndAttackers, (heuristic, attackers))
            # this heap that is printed (if uncommented) shows the best heuristic for the blocking player
            # given a certain list of attackers. Therefore, the MCTS player who is attacking will choose 
            # the list of attackers that corresponds to the min of the max of the blockers heuristic
            # print "ENTIRE HEAP " + str(heuristicsAndAttackers) + "\n"
            return heappop(heuristicsAndAttackers)[1]
        else:
            return []

    # similar to the function above, this function returns the state before and after battle
    # for a random player
    def RandomAttack(self, primary, opponent):
        attackList = self.board.getPlayer(primary).getAttackList()
        print "Before combat:\n"
        print self.board
        print "Combat just happened. Here's the result:\n"
        if len(attackList) > 0:
            return random.choice(attackList)
        else:
            return []
        
    # this function handles turns of all random players
    def RandomPlay(self, primary, opponent):
        newBoard = copy.deepcopy(self.board)
        primaryPlayer = newBoard.getPlayer(primary)
        opponentPlayer = newBoard.getPlayer(opponent)
        # play random creatureList:::
        powerSetCreatures = primaryPlayer.getPlayableCreatures()
        creatureList = []
        if len(powerSetCreatures) > 0:
            creatureList = random.choice(powerSetCreatures)
        for creature in creatureList:
            primaryPlayer.playCreature(creature)
        return newBoard

    # this function gets the board with the highest heuristic given a list of attackers

    def MCTSBlock(self, primary, opponent, attackers):
        boards = self.getPossibleBoardsForAttackers(opponent, primary, attackers)
        heuristicsAndBoards = []
        [board, heuristic] = self.getHeuristicMCTS(opponent, primary, boards)
        return board
        
    # this function gets a board after blocking with a random set of creatures
    def RandomBlock(self, primary, opponent, attackers):
        boards = self.getPossibleBoardsForAttackers(opponent, primary, attackers)
        if len(boards) > 0:
            return random.choice(boards)
        else:
            return self.board

    # this function handles the turns of both random players
    # and MCTS AI players and handles each turn according to the type of each player
    def doAI(self, primary, opponent):
        if self.board.getPlayer(primary).getType() == TYPEMCTS:
            attackers = self.MCTSPlayAndAttack(primary, opponent)
        elif self.board.getPlayer(primary).getType() == TYPERANDOM:
            attackers = self.RandomPlayAndAttack(primary, opponent)
        else:
            print "FATAL ERROR"

        if len(attackers) > 0:
            if self.board.getPlayer(opponent).getType() == TYPEMCTS:
                self.board = self.MCTSBlock(opponent, primary, attackers)
            elif self.board.getPlayer(opponent).getType() == TYPERANDOM:
                self.board = self.RandomBlock(opponent, primary, attackers)
            else:
                print "FATAL ERROR"

    # this function is used only in testing quality of the decks we created
    # by doing random player for a as many rounds (1000 rounds was as high as we could 
    # reasonably go with our computing power) we were able to see how evenly matched
    # each deck was so we could hypothesize the outcome of our AI player
    def randomPlaySimulation(self, primary, opponent):
        primaryPlayer = self.board.getPlayer(primary)
        opponentPlayer = self.board.getPlayer(opponent)
        playingList = primaryPlayer.quasiPlayList()
        for i in playingList:
            primaryPlayer.playCreature(i)
        attackList = primaryPlayer.quasiAttackList()
        for d in attackList:
            blocker = opponentPlayer.quasiBlockersInPlay(d)
            self.doBattle(self.board, primary, d, opponent, blocker, True)

    # this function gets the number of wins of a player during the random simulations
    # which are part of MCTS
    def checkRandomOver(self, player1, player2):
        global gameOver       
        if self.board.getPlayer(player1).isDead():
            gameOverMessage = "Player " + self.board.getPlayer(0).getName() + " lost"
            gameOver = True
            self.board.getPlayer(player2).incrementWinCount()
        elif self.board.getPlayer(player2).isDead():
            gameOverMessage = "Player " + self.board.getPlayer(1).getName() + " lost"
            gameOver = True
            self.board.getPlayer(player1).incrementWinCount()

    # for each of the expanded board states, a heuristic is run by playing 
    # a number of games that is determined by the 'rounds' variable. 
    # the board state that wins the most for the current player is returned
    # where the different board states are a result of the different ways to 
    # play creatures and then do battle (attack)
    def getHeuristicMCTS(self, primary, opponent, boards):
        global gameOver
        global mainPlayList
        global attackList
        global primaryHeap
        global opponentHeap
        global opponentDictionary
        bestBoard = None
        bestHeuristic = -99999
        p1CopyWinCount = 0
        p2CopyWinCount = 0
        for board in boards:
            p1CopyWinCount = 0
            p2CopyWinCount = 0
            rounds = 20
            board.getPlayer(primary).clearWinCount()
            board.getPlayer(opponent).clearWinCount()
            while (rounds > 0):
                p1Copy = copy.deepcopy(board.getPlayer(primary))
                p2Copy = copy.deepcopy(board.getPlayer(opponent))
                gameOver = False
                mainPlayList = []
                attackList = []
                primaryHeap = []
                opponentHeap = []
                opponentDictionary = {}
                rounds -= 1
                p1Copy = copy.deepcopy(board.getPlayer(primary))
                p2Copy = copy.deepcopy(board.getPlayer(opponent))
                if primary == 0:
                    magicGame = Game(p1Copy, p2Copy)
                else:
                    magicGame = Game(p2Copy, p1Copy)
                magicGame.activePlayer = self.activePlayer
                while(not gameOver):
                    # print "Turn: " + str(magicGame.turn)
                    magicGame.doRandomTurns()
                    magicGame.checkRandomOver(primary, opponent)
                    magicGame.advanceTurn()
                if p1Copy.getWinCount() == 1:
                    p1CopyWinCount += 1
                elif p2Copy.getWinCount() == 1:
                    p2CopyWinCount += 1
            if (p2CopyWinCount > bestHeuristic):
                bestBoard = board
                bestHeuristic = p2CopyWinCount 
        gameOver = False
        return [bestBoard, bestHeuristic]

    # this handles the setup and cleanup of the random simulations
    def doRandomTurns(self):
        current = self.board.getPlayer(self.activePlayer)
        current.clearAll()
        if self.turn != 0:
            current.draw()
        # print "Land step"
        for card in current.getHand():
            if card.isLand():
                current.playLand(card)
                break
        self.randomPlaySimulation(self.activePlayer, (self.activePlayer + 1) % 2)
        current.endTurn()

 
def main():
    global gameOver
    global player1WinCount 
    global player2WinCount
    global gameOver
    global mainPlayList
    global attackList
    global primaryHeap
    global opponentHeap
    global opponentDictionary
    global turn
    print "Initializing..."
    player1WinCount = 0
    player2WinCount = 0
    initializeDecks()
    print "Enter deck choice by number:"
    for i in range(0,len(decks)):
        print "Deck " + str(i) + ": " + decks[i].getName()
    # choose which deck you would like to play
    player1Input = int(raw_input("Which deck is player 1 using? 0 or 1 "))
    player2Input = int(raw_input("Which deck is player 2 using? 0 or 1 "))
    
    player1Deck = copy.deepcopy(decks[player1Input])
    player2Deck = copy.deepcopy(decks[player2Input])
    player1 = Player(player1Deck, TYPEMCTS)
    player2 = Player(player2Deck, TYPEMCTS)
    magicGame = Game(player1, player2)
    while(not gameOver):
            print "Turn: " + str(magicGame.turn)
            magicGame.doTurn()
            magicGame.checkOver()
            magicGame.advanceTurn()
            print "this may take a long time depending on the current board state"
            input = raw_input("press enter to continue simulation. press ctrl-d to quit")

    print "SIMULATION OVER"
    
# this function creates the different decks we chose to use in our AI. 
# the possibilities are boundless so we chose to recreate classic 
# deck archetypes that have been around since the beggining of 
# Magic The Gathering. 

# White Weenie is a deck that has a large number of 'small' creatures
# with low casting costs which overwhelms the opponent with the early
# start

# Big Green is a deck that has a modest number of 'large' creatures
# with large casting costs that may be slow to start, but has the 
# potential to play very powerful creatures in the late game
def initializeDecks():
    for i in range(1, 10):
        for j in range(1, 10):
            for color in COLORS:
                name = color + "Vanilla" + str(i) + str(j)
                cost = []
                for k in range(0, (i + j) / 2):
                    cost.append(color)
                card = Creature(name, cost, [], [color], i, j)
                dictOfCards[name] = card
    #Format is "GVanilla23" for a green vanilla 2/3 creature

    plains = Land("Plains", WHITE)
    mountain = Land("Mountain", RED)
    swamp = Land("Swamp", BLACK)
    island = Land("Island", BLUE)
    forest = Land("Forest", GREEN)

    whiteWeenie = Deck("White Weenie")
    decks.append(whiteWeenie)
    whiteWeenie.addCard(plains, 24)
    whiteWeenie.addCard(dictOfCards[WHITE + "Vanilla" + str(1) + str(1)], 4)
    whiteWeenie.addCard(dictOfCards[WHITE + "Vanilla" + str(2) + str(2)], 4)
    whiteWeenie.addCard(dictOfCards[WHITE + "Vanilla" + str(3) + str(3)], 4)
    whiteWeenie.addCard(dictOfCards[WHITE + "Vanilla" + str(4) + str(4)], 4)
    whiteWeenie.addCard(dictOfCards[WHITE + "Vanilla" + str(5) + str(5)], 3)
    whiteWeenie.addCard(dictOfCards[WHITE + "Vanilla" + str(2) + str(1)], 4)
    whiteWeenie.addCard(dictOfCards[WHITE + "Vanilla" + str(3) + str(2)], 4)
    whiteWeenie.addCard(dictOfCards[WHITE + "Vanilla" + str(4) + str(2)], 4)
    whiteWeenie.addCard(dictOfCards[WHITE + "Vanilla" + str(6) + str(6)], 2)
    whiteWeenie.addCard(dictOfCards[WHITE + "Vanilla" + str(1) + str(4)], 3)

    whiteWeenie2 = Deck("White Weenie2")
    decks.append(whiteWeenie2)
    whiteWeenie2.addCard(plains, 24)
    whiteWeenie2.addCard(dictOfCards[WHITE + "Vanilla" + str(1) + str(1)], 4)
    whiteWeenie2.addCard(dictOfCards[WHITE + "Vanilla" + str(2) + str(2)], 4)
    whiteWeenie2.addCard(dictOfCards[WHITE + "Vanilla" + str(3) + str(3)], 4)
    whiteWeenie2.addCard(dictOfCards[WHITE + "Vanilla" + str(4) + str(4)], 4)
    whiteWeenie2.addCard(dictOfCards[WHITE + "Vanilla" + str(5) + str(5)], 3)
    whiteWeenie2.addCard(dictOfCards[WHITE + "Vanilla" + str(2) + str(1)], 4)
    whiteWeenie2.addCard(dictOfCards[WHITE + "Vanilla" + str(3) + str(2)], 4)
    whiteWeenie2.addCard(dictOfCards[WHITE + "Vanilla" + str(4) + str(2)], 4)
    whiteWeenie2.addCard(dictOfCards[WHITE + "Vanilla" + str(6) + str(6)], 2)
    whiteWeenie2.addCard(dictOfCards[WHITE + "Vanilla" + str(1) + str(4)], 3)

    bigGreen = Deck("Big Green")
    decks.append(bigGreen)
    bigGreen.addCard(forest, 26)
    bigGreen.addCard(dictOfCards[GREEN + "Vanilla" + str(1) + str(1)], 4)
    bigGreen.addCard(dictOfCards[GREEN + "Vanilla" + str(2) + str(2)], 4)
    bigGreen.addCard(dictOfCards[GREEN + "Vanilla" + str(3) + str(3)], 4)
    bigGreen.addCard(dictOfCards[GREEN + "Vanilla" + str(4) + str(4)], 4)
    bigGreen.addCard(dictOfCards[GREEN + "Vanilla" + str(5) + str(5)], 4)
    bigGreen.addCard(dictOfCards[GREEN + "Vanilla" + str(6) + str(6)], 4)
    bigGreen.addCard(dictOfCards[GREEN + "Vanilla" + str(7) + str(7)], 4)
    bigGreen.addCard(dictOfCards[GREEN + "Vanilla" + str(8) + str(8)], 4)
    bigGreen.addCard(dictOfCards[GREEN + "Vanilla" + str(9) + str(9)], 2)

main()
